# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import dataclasses
import os
import pathlib
import subprocess  # nosec B404
from typing import TextIO
import uuid

from ansys.tools.local_product_launcher.grpc_transport import (
    InsecureOptions,
    MTLSOptions,
    TransportOptionsType,
    UDSOptions,
    WNUAOptions,
)
from ansys.tools.local_product_launcher.helpers.grpc import check_grpc_health
from ansys.tools.local_product_launcher.helpers.ports import find_free_ports
from ansys.tools.local_product_launcher.interface import (
    METADATA_KEY_DOC,
    LauncherProtocol,
    ServerType,
)
from ansys.tools.path import get_latest_ansys_installation

from .common import ServerKey

__all__ = ["DirectLaunchConfig"]

_GRPC_MAX_MESSAGE_LENGTH = 256 * 1024**2  # 256 MB


def _get_default_binary_path() -> str:
    try:
        _, ans_root = get_latest_ansys_installation()
        binary_path = os.path.join(ans_root, "ACP", "acp_grpcserver")
        if os.name == "nt":
            binary_path += ".exe"
        return binary_path
    except (ValueError, FileNotFoundError):
        return ""


@dataclasses.dataclass
class DirectLaunchConfig:
    """Configuration options for launching ACP as a sub-process."""

    binary_path: str = dataclasses.field(
        default=_get_default_binary_path(),
        metadata={METADATA_KEY_DOC: "Path to the ACP gRPC server executable."},
    )
    """Path to the ACP gRPC server executable."""

    stdout_file: str = dataclasses.field(
        default=os.devnull,
        metadata={METADATA_KEY_DOC: "File in which the server stdout is stored."},
    )
    """File in which the server stdout is stored."""

    stderr_file: str = dataclasses.field(
        default=os.devnull,
        metadata={METADATA_KEY_DOC: "File in which the server stderr is stored."},
    )
    """File in which the server stderr is stored."""

    transport_mode: str = dataclasses.field(
        default="wnua" if os.name == "nt" else "uds",
        metadata={
            METADATA_KEY_DOC: "Specifies the gRPC transport mode to use. "
            f"Possible values: '{'wnua' if os.name == 'nt' else 'uds'}' (default), 'mtls', 'insecure'."
        },
    )
    """Specifies the gRPC transport mode to use.

    Possible values are:

    - ``"uds"`` : Unix Domain Sockets (default on Unix systems, unsupported on Windows)
    - ``"wnua"`` : Windows Named User Authentication (default on Windows systems, unsupported on Unix)
    - ``"mtls"`` : Mutual TLS
    - ``"insecure"`` : Insecure TCP connection (not recommended)
    """

    uds_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Directory path for UDS socket files (default: ~/.conn). "
            "Only used if transport_mode is 'uds'."
        },
    )
    """Directory path for UDS socket files.

    Defaults to ``~/.conn``. Only used if ``transport_mode`` is ``"uds"``.
    """

    certs_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Directory path for mTLS certificate files. Only used if transport_mode is 'mtls'."
        },
    )
    """Directory path for mTLS certificate files.

    Defaults to the ``ANSYS_GRPC_CERTIFICATES`` environment variable, or ``certs`` if the variable is not set.
    Only used if ``transport_mode`` is ``"mtls"``.
    """


class DirectLauncher(LauncherProtocol[DirectLaunchConfig]):
    CONFIG_MODEL = DirectLaunchConfig
    SERVER_SPEC = {ServerKey.MAIN: ServerType.GRPC}

    def __init__(self, *, config: DirectLaunchConfig):
        self._config = config
        self._process: subprocess.Popen[str] | None = None
        self._stdout: TextIO
        self._stderr: TextIO
        self._url: str
        self._transport_options: TransportOptionsType

    def start(self) -> None:
        stdout_file = self._config.stdout_file
        stderr_file = self._config.stderr_file

        # Sanity checks
        binary = pathlib.Path(self._config.binary_path)
        if not binary.exists():
            raise FileNotFoundError(f"Binary not found: '{binary}'")
        if os.name == "nt":
            if self._config.transport_mode == "uds":
                raise RuntimeError("UDS transport mode is not supported on Windows.")
        else:
            if self._config.transport_mode == "wnua":
                raise RuntimeError("WNUA transport mode is only supported on Windows.")
        # Determine if the patched or unpatched version of the server is used
        is_patched_server = (
            "allow-remote-host"
            in subprocess.check_output(  # nosec B603, B607: documented in 'security_considerations.rst'
                [
                    self._config.binary_path,
                    "--help",
                ],
                text=True,
            )
        )
        if not is_patched_server and self._config.transport_mode != "insecure":
            raise RuntimeError(
                f"The {self._config.transport_mode} transport mode requires a patched version "
                "of the ACP gRPC server. Install the latest Service Pack for your "
                "Ansys installation."
            )

        if self._config.transport_mode == "uds":
            uds_id = uuid.uuid4().hex
            transport_args = [
                "--transport-mode=uds",
                f"--uds-id={uds_id}",
            ]
            if self._config.uds_dir is not None:
                transport_args.append(f"--uds-dir={self._config.uds_dir}")

            self._transport_options = UDSOptions(
                uds_service="acp_grpcserver",
                uds_dir=self._config.uds_dir,
                uds_id=uds_id,
            )
        else:
            port = find_free_ports()[0]
            if self._config.transport_mode == "wnua":
                transport_args = [
                    "--transport-mode=wnua",
                    f"--port={port}",
                ]
                self._transport_options = WNUAOptions(port=port)
            elif self._config.transport_mode == "mtls":
                if self._config.certs_dir is None:
                    raise ValueError("certs_dir must be specified for mtls transport mode.")
                transport_args = [
                    "--transport-mode=mtls",
                    "--host=localhost",
                    f"--port={port}",
                ]
                if self._config.certs_dir is not None:
                    transport_args.extend(["--certs-dir", str(self._config.certs_dir)])
                self._transport_options = MTLSOptions(
                    certs_dir=self._config.certs_dir,
                    host="localhost",
                    port=port,
                    allow_remote_host=False,
                )
            elif self._config.transport_mode == "insecure":
                if is_patched_server:
                    transport_args = [
                        "--transport-mode=insecure",
                        "--host=localhost",
                        f"--port={port}",
                    ]
                else:
                    transport_args = [
                        f"--server-address=localhost:{port}",
                    ]

                self._transport_options = InsecureOptions(
                    host="localhost",
                    port=port,
                )

        self._stdout = open(stdout_file, mode="w", encoding="utf-8")
        self._stderr = open(stderr_file, mode="w", encoding="utf-8")
        self._process = subprocess.Popen(  # nosec B603: documented in 'security_considerations.rst'
            [
                self._config.binary_path,
            ]
            + transport_args,
            stdout=self._stdout,
            stderr=self._stderr,
            text=True,
        )

    def stop(self, *, timeout: float | None = None) -> None:
        if self._process is None:
            # The process has not been started, and therefore doesn't need to be stopped
            return
        self._process.terminate()
        try:
            self._process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            self._process.kill()
            self._process.wait()
        self._stdout.close()
        self._stderr.close()

    def check(self, timeout: float | None = None) -> bool:
        channel = self._transport_options.create_channel()
        return check_grpc_health(channel=channel, timeout=timeout)

    @property
    def transport_options(self) -> dict[str, TransportOptionsType]:
        return {ServerKey.MAIN: self._transport_options}



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections.abc import Callable
from functools import wraps
from typing import ParamSpec, TypeVar

P = ParamSpec("P")
T = TypeVar("T")


def requires_pyvista(func: Callable[P, T]) -> Callable[P, T]:
    """Decorate a function as requiring the 'pyvista' package.

    Checks if the 'pyvista' package is installed before executing the
    decorated function, and provides a helpful error message if it is not.
    """

    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            import pyvista  # noqa
        except ImportError as exc:
            raise ImportError(
                f"The '{func.__name__}' function requires the 'pyvista' package. "
                "Please reinstall PyACP with 'pip install ansys-acp-core[graphics]'."
            ) from exc
        return func(*args, **kwargs)

    return inner



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import dataclasses
import pathlib

from ansys.tools.local_product_launcher.grpc_transport import (
    InsecureOptions,
    MTLSOptions,
    TransportOptionsType,
    UDSOptions,
    WNUAOptions,
)
from ansys.tools.local_product_launcher.helpers.grpc import check_grpc_health
from ansys.tools.local_product_launcher.interface import (
    METADATA_KEY_DOC,
    LauncherProtocol,
    ServerType,
)

from .common import ServerKey

__all__ = ["ConnectLaunchConfig", "ConnectLocalLaunchConfig"]


@dataclasses.dataclass(kw_only=True)
class ConnectLaunchConfig:
    """Configuration options for attaching to an existing ACP server."""

    acp_transport_mode: str = dataclasses.field(
        default="mtls",
        metadata={
            METADATA_KEY_DOC: "Specifies the gRPC transport mode to use for the main ACP server. "
            "Possible values: 'mtls' (default), 'uds' (Unix only), 'wnua' (Windows only), 'insecure'."
        },
    )
    """Specifies the gRPC transport mode to use for the main ACP server.

    Possible values are:

    - ``"mtls"`` : Mutual TLS (default)
    - ``"uds"`` : Unix Domain Sockets (unsupported on Windows)
    - ``"wnua"`` : Windows Named User Authentication (unsupported on Unix)
    - ``"insecure"`` : Insecure TCP connection (not recommended)
    """

    acp_host: str = dataclasses.field(
        default="localhost",
        metadata={
            METADATA_KEY_DOC: "Hostname or IP address to connect to for the main ACP server."
        },
    )
    """Hostname or IP address to connect to for the main ACP server."""

    acp_port: int = dataclasses.field(
        default=50555,
        metadata={METADATA_KEY_DOC: "Port number to connect to for the main ACP server."},
    )
    """Port number to connect to for the main ACP server."""

    acp_uds_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: (
                "Directory path for UDS socket files used by the main ACP server (default: ~/.conn). "
                "Only used if acp_transport_mode is 'uds'."
            )
        },
    )
    """
    Directory path for UDS socket files used by the main ACP server.

    Defaults to ``~/.conn``. Only used if ``acp_transport_mode`` is ``"uds"``.
    """

    acp_uds_id: str | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Optional ID for UDS socket file naming for the main ACP server "
            "(acp_grpcserver-<id>.sock). Only used if acp_transport_mode is 'uds'."
        },
    )
    """
    Optional ID for UDS socket file naming for the main ACP server (acp_grpcserver-<id>.sock).
    Only used if ``acp_transport_mode`` is ``"uds"``.
    """

    acp_certs_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Directory path for mTLS certificate files for the main ACP server. "
            "Only used if acp_transport_mode is 'mtls'."
        },
    )
    """Directory path for mTLS certificate files for the main ACP server.

    Defaults to the ``ANSYS_GRPC_CERTIFICATES`` environment variable, or ``certs`` if the variable is not set.
    Only used if ``acp_transport_mode`` is ``"mtls"``.
    """

    acp_allow_remote_host: bool = dataclasses.field(
        default=False,
        metadata={
            METADATA_KEY_DOC: "Whether to allow connecting to a remote host for the main ACP server."
        },
    )
    """Whether to allow connecting to a remote host for the main ACP server."""

    filetransfer_transport_mode: str = dataclasses.field(
        default="mtls",
        metadata={
            METADATA_KEY_DOC: "Specifies the gRPC transport mode to use. "
            "Possible values: 'mtls' (default), 'uds', 'insecure'."
        },
    )
    """Specifies the gRPC transport mode to use for the file transfer server.

    Possible values are:

    - ``"mtls"`` : Mutual TLS (default)
    - ``"uds"`` : Unix Domain Sockets (unsupported on Windows)
    - ``"insecure"`` : Insecure TCP connection (not recommended)
    """

    filetransfer_host: str = dataclasses.field(
        default="localhost",
        metadata={METADATA_KEY_DOC: "Host to connect to for the file transfer server."},
    )
    """Hostname or IP address to connect to for the file transfer server."""

    filetransfer_port: int = dataclasses.field(
        default=50556,
        metadata={METADATA_KEY_DOC: "Port to connect to for the file transfer server."},
    )
    """Port number to connect to for the file transfer server."""

    filetransfer_uds_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: (
                "Directory path for UDS socket files used by the filetransfer server (default: ~/.conn). "
                "Only used if acp_transport_mode is 'uds'."
            )
        },
    )
    """
    Directory path for UDS socket files used by the filetransfer server.

    Defaults to ``~/.conn``. Only used if ``filetransfer_transport_mode`` is ``"uds"``.
    """

    filetransfer_uds_id: str | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Optional ID for UDS socket file naming for the filetransfer server "
            "(filetransfer_grpcserver-<id>.sock). Only used if filetransfer_transport_mode is 'uds'."
        },
    )
    """
    Optional ID for UDS socket file naming for the filetransfer server (filetransfer_grpcserver-<id>.sock).
    Only used if ``filetransfer_transport_mode`` is ``"uds"``.
    """

    filetransfer_certs_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Directory containing TLS certificates for file transfer. "
            "Only used if filetransfer_transport_mode is 'mtls'."
        },
    )
    """Directory path for mTLS certificate files for the filetransfer server.

    Defaults to the ``ANSYS_GRPC_CERTIFICATES`` environment variable, or ``certs`` if the variable is not set.
    Only used if ``filetransfer_transport_mode`` is ``"mtls"``.
    """

    filetransfer_allow_remote_host: bool = dataclasses.field(
        default=False,
        metadata={
            METADATA_KEY_DOC: "Whether to allow connecting to a remote host for the file transfer server."
        },
    )
    """Whether to allow connecting to a remote host for the filetransfer server."""


class ConnectLauncher(LauncherProtocol[ConnectLaunchConfig]):
    CONFIG_MODEL = ConnectLaunchConfig
    SERVER_SPEC = {ServerKey.MAIN: ServerType.GRPC, ServerKey.FILE_TRANSFER: ServerType.GRPC}

    def __init__(self, *, config: ConnectLaunchConfig):
        self._config = config

        if self._config.acp_transport_mode == "uds":
            acp_transport_options: TransportOptionsType = UDSOptions(
                uds_service="acp_grpcserver",
                uds_dir=self._config.acp_uds_dir,
                uds_id=self._config.acp_uds_id,
            )
        elif self._config.acp_transport_mode == "wnua":
            acp_transport_options = WNUAOptions(
                port=self._config.acp_port,
            )
        elif self._config.acp_transport_mode == "mtls":
            acp_transport_options = MTLSOptions(
                certs_dir=self._config.acp_certs_dir,
                host=self._config.acp_host,
                port=self._config.acp_port,
                allow_remote_host=self._config.acp_allow_remote_host,
            )
        elif self._config.acp_transport_mode == "insecure":
            acp_transport_options = InsecureOptions(
                host=self._config.acp_host,
                port=self._config.acp_port,
                allow_remote_host=self._config.acp_allow_remote_host,
            )
        else:
            raise ValueError(f"Invalid transport mode for ACP: {self._config.acp_transport_mode}")

        if self._config.filetransfer_transport_mode == "uds":
            filetransfer_transport_options: UDSOptions | MTLSOptions | InsecureOptions = UDSOptions(
                uds_service="filetransfer_grpcserver",
                uds_dir=self._config.filetransfer_uds_dir,
                uds_id=self._config.filetransfer_uds_id,
            )
        elif self._config.filetransfer_transport_mode == "mtls":
            filetransfer_transport_options = MTLSOptions(
                certs_dir=self._config.filetransfer_certs_dir,
                host=self._config.filetransfer_host,
                port=self._config.filetransfer_port,
                allow_remote_host=self._config.filetransfer_allow_remote_host,
            )
        elif self._config.filetransfer_transport_mode == "insecure":
            filetransfer_transport_options = InsecureOptions(
                host=self._config.filetransfer_host,
                port=self._config.filetransfer_port,
                allow_remote_host=self._config.filetransfer_allow_remote_host,
            )
        else:
            raise ValueError(
                f"Invalid transport mode for filetransfer: {self._config.filetransfer_transport_mode}"
            )

        self._acp_transport_options = acp_transport_options
        self._filetransfer_transport_options = filetransfer_transport_options

    def start(self) -> None:
        # Since this launcher simply connects to an existing server, we don't need to start it.
        return

    def stop(self, *, timeout: float | None = None) -> None:
        # Since this launcher simply connects to an existing server, we don't need to stop it.
        return

    def check(self, timeout: float | None = None) -> bool:
        for transport_opt in self.transport_options.values():
            channel = transport_opt.create_channel()
            if not check_grpc_health(channel=channel, timeout=timeout):
                return False
        return True

    @property
    def transport_options(self) -> dict[str, TransportOptionsType]:
        return {
            ServerKey.MAIN: self._acp_transport_options,
            ServerKey.FILE_TRANSFER: self._filetransfer_transport_options,
        }


@dataclasses.dataclass
class ConnectLocalLaunchConfig:
    """Configuration options for attaching to an existing ACP server without filetransfer."""

    transport_mode: str = dataclasses.field(
        default="mtls",
        metadata={
            METADATA_KEY_DOC: "Specifies the gRPC transport mode to use."
            "Possible values: 'mtls' (default), 'uds' (Unix only), 'wnua' (Windows only), 'insecure'."
        },
    )
    """Specifies the gRPC transport mode to use.

    Possible values are:
    - ``"mtls"`` : Mutual TLS (default)
    - ``"uds"`` : Unix Domain Sockets (unsupported on Windows)
    - ``"wnua"`` : Windows Named User Authentication (unsupported on Unix)
    - ``"insecure"`` : Insecure TCP connection (not recommended)
    """

    host: str = dataclasses.field(
        default="localhost",
        metadata={METADATA_KEY_DOC: "Hostname or IP to connect to."},
    )
    """Hostname or IP to connect to."""

    port: int = dataclasses.field(
        default=50555,
        metadata={METADATA_KEY_DOC: "Port number to connect to."},
    )
    """Port number to connect to."""

    uds_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: (
                "Directory path for UDS socket files (default: ~/.conn). "
                "Only used if transport_mode is 'uds'."
            )
        },
    )
    """Directory path for UDS socket files.

    Defaults to ``~/.conn``. Only used if ``transport_mode`` is ``"uds"``.
    """

    uds_id: str | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Optional ID for UDS socket file naming (acp_grpcserver-<id>.sock). "
            "Only used if transport_mode is 'uds'."
        },
    )
    """
    Optional ID for UDS socket file naming (acp_grpcserver-<id>.sock).
    Only used if ``transport_mode`` is ``"uds"``.
    """

    certs_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Directory path for mTLS certificate files. "
            "Only used if transport_mode is 'mtls'."
        },
    )
    """Directory path for mTLS certificate files.

    Defaults to the ``ANSYS_GRPC_CERTIFICATES`` environment variable, or ``certs`` if the variable is not set.
    Only used if ``transport_mode`` is ``"mtls"``.
    """

    allow_remote_host: bool = dataclasses.field(
        default=False,
        metadata={METADATA_KEY_DOC: "Whether to allow connecting to a remote host."},
    )
    """Whether to allow connecting to a remote host."""


class ConnectLocalLauncher(LauncherProtocol[ConnectLocalLaunchConfig]):
    CONFIG_MODEL = ConnectLocalLaunchConfig
    SERVER_SPEC = {ServerKey.MAIN: ServerType.GRPC}

    def __init__(self, *, config: ConnectLocalLaunchConfig):
        self._config = config
        if self._config.transport_mode == "uds":
            acp_transport_options: TransportOptionsType = UDSOptions(
                uds_service="acp_grpcserver",
                uds_dir=self._config.uds_dir,
                uds_id=self._config.uds_id,
            )
        elif self._config.transport_mode == "wnua":
            acp_transport_options = WNUAOptions(
                port=self._config.port,
            )
        elif self._config.transport_mode == "mtls":
            acp_transport_options = MTLSOptions(
                certs_dir=self._config.certs_dir,
                host=self._config.host,
                port=self._config.port,
                allow_remote_host=self._config.allow_remote_host,
            )
        elif self._config.transport_mode == "insecure":
            acp_transport_options = InsecureOptions(
                host=self._config.host,
                port=self._config.port,
                allow_remote_host=self._config.allow_remote_host,
            )
        else:
            raise ValueError(f"Invalid transport mode for ACP: {self._config.transport_mode}")

        self._acp_transport_options = acp_transport_options

    def start(self) -> None:
        # Since this launcher simply connects to an existing server, we don't need to start it.
        return

    def stop(self, *, timeout: float | None = None) -> None:
        # Since this launcher simply connects to an existing server, we don't need to stop it.
        return

    def check(self, timeout: float | None = None) -> bool:
        for transport_opt in self.transport_options.values():
            channel = transport_opt.create_channel()
            if not check_grpc_health(channel=channel, timeout=timeout):
                return False
        return True

    @property
    def transport_options(self) -> dict[str, TransportOptionsType]:
        return {
            ServerKey.MAIN: self._acp_transport_options,
        }



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
from collections.abc import Iterator
import contextlib
import copy
import dataclasses
import importlib.resources
import math
import os
import pathlib
import subprocess  # nosec B404
import uuid

from packaging.version import parse as parse_version

from ansys.tools.local_product_launcher.grpc_transport import (
    InsecureOptions,
    MTLSOptions,
)
from ansys.tools.local_product_launcher.helpers.grpc import check_grpc_health
from ansys.tools.local_product_launcher.helpers.ports import find_free_ports
from ansys.tools.local_product_launcher.interface import (
    METADATA_KEY_DOC,
    METADATA_KEY_NOPROMPT,
    LauncherProtocol,
    ServerType,
)

from .common import ServerKey

__all__ = ["DockerComposeLaunchConfig"]

TransportOptionsType = (
    InsecureOptions | MTLSOptions
)  # UDS and WNUA are not supported for docker-compose


def _get_default_license_server() -> str:
    try:
        return os.environ["ANSYSLMD_LICENSE_FILE"]
    except KeyError:
        return ""


@dataclasses.dataclass
class DockerComposeLaunchConfig:
    """Configuration options for launching ACP through docker compose."""

    image_name_acp: str = dataclasses.field(
        default="ghcr.io/ansys/acp:latest",
        metadata={METADATA_KEY_DOC: "Docker image running the ACP gRPC server."},
    )
    """Docker image running the ACP gRPC server."""

    image_name_filetransfer: str = dataclasses.field(
        default="ghcr.io/ansys/tools-filetransfer:latest",
        metadata={METADATA_KEY_DOC: "Docker image running the file transfer service."},
    )
    """Docker image running the file transfer service."""

    license_server: str = dataclasses.field(
        default=_get_default_license_server(),
        metadata={
            METADATA_KEY_DOC: (
                "License server passed to the container as "
                "'ANSYSLMD_LICENSE_FILE' environment variable."
            )
        },
    )
    """License server passed to the container as 'ANSYSLMD_LICENSE_FILE' environment variable."""

    keep_volume: bool = dataclasses.field(
        default=False,
        metadata={METADATA_KEY_DOC: "If true, keep the volume after docker compose is stopped."},
    )
    """If true, keep the volume after docker compose is stopped."""

    compose_file: str | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: (
                "Docker compose file used to start the services. Uses the "
                "compose files shipped with PyACP by default."
            ),
            METADATA_KEY_NOPROMPT: True,
        },
    )
    """Docker compose file used to start the services. Uses the compose files shipped with PyACP by default."""

    environment_variables: dict[str, str] = dataclasses.field(
        default_factory=dict,
        metadata={
            METADATA_KEY_DOC: (
                "Additional environment variables passed to docker compose. These take "
                "precedence over environment variables defined through another configuration "
                "option (for example 'license_server' which defines 'ANSYSLMD_LICENSE_FILE') "
                "or the pre-existing environment variables."
            ),
            METADATA_KEY_NOPROMPT: True,
        },
    )
    """Additional environment variables passed to docker compose.

    These take precedence over environment variables defined through another configuration
    option (for example 'license_server' which defines 'ANSYSLMD_LICENSE_FILE') or the pre-existing
    environment variables.
    """

    transport_mode: str = dataclasses.field(
        default="mtls",
        metadata={
            METADATA_KEY_DOC: "Specifies the gRPC transport mode to use. Only 'mtls' and 'insecure' are supported."
        },
    )
    """Specifies the gRPC transport mode to use.

    Possible values are:

    - ``"mtls"`` : Mutual TLS
    - ``"insecure"`` : Insecure TCP connection (not recommended)
    """

    certs_dir: str | pathlib.Path | None = dataclasses.field(
        default=None,
        metadata={
            METADATA_KEY_DOC: "Directory containing TLS certificates. Only used if transport_mode is 'mtls'."
        },
    )
    """Directory path for mTLS certificate files.

    Defaults to the ``ANSYS_GRPC_CERTIFICATES`` environment variable, or ``certs`` if the variable is not set.
    Only used if ``transport_mode`` is ``"mtls"``.
    """


class DockerComposeLauncher(LauncherProtocol[DockerComposeLaunchConfig]):
    CONFIG_MODEL = DockerComposeLaunchConfig
    SERVER_SPEC = {ServerKey.MAIN: ServerType.GRPC, ServerKey.FILE_TRANSFER: ServerType.GRPC}

    def __init__(self, *, config: DockerComposeLaunchConfig):
        self._compose_name = f"pyacp_compose_{uuid.uuid4().hex}"
        self._config = config
        self._transport_options: dict[str, TransportOptionsType] = {}

        try:
            import ansys.tools.filetransfer  # noqa
        except ImportError as err:
            raise ImportError(
                "The 'ansys.tools.filetransfer' module is needed to launch ACP via docker-compose."
            ) from err

        if self._config.transport_mode == "mtls" and self._config.certs_dir is None:
            # The default 'certs_dir' needs to be implemented explicitly here
            # since it needs to be mounted into the docker container.
            self._config.certs_dir = os.environ.get("ANSYS_GRPC_CERTIFICATES")
            if self._config.certs_dir is None:
                self._config.certs_dir = pathlib.Path.cwd() / "certs"

        self._env = copy.deepcopy(os.environ)
        self._env.update(
            IMAGE_NAME_ACP=self._config.image_name_acp,
            IMAGE_NAME_FILETRANSFER=self._config.image_name_filetransfer,
            ANSYSLMD_LICENSE_FILE=self._config.license_server,
        )
        self._env.update(self._config.environment_variables)
        if self._config.transport_mode == "mtls":
            assert self._config.certs_dir is not None
            self._env["CERTS_DIR"] = str(pathlib.Path(self._config.certs_dir).resolve())

        self._keep_volume = config.keep_volume

        if config.compose_file is not None:
            self._compose_file: pathlib.Path | None = pathlib.Path(config.compose_file)
        else:
            self._compose_file = None

        try:
            self._compose_version = parse_version(
                subprocess.check_output(  # nosec B603, B607: documented in 'security_considerations.rst'
                    ["docker", "compose", "version", "--short"], text=True
                ).replace(
                    "-", "+"
                )
            )
            self._compose_cmds = ["docker", "compose"]
        except subprocess.CalledProcessError:
            # If 'docker compose does not work, try 'docker-compose' instead.
            self._compose_version = parse_version(
                subprocess.check_output(  # nosec B603, B607: documented in 'security_considerations.rst'
                    ["docker-compose", "version", "--short"], text=True
                ).replace(
                    "-", "+"
                )
            )
            self._compose_cmds = ["docker-compose"]

    @contextlib.contextmanager
    def _get_compose_file(self) -> Iterator[pathlib.Path]:
        if self._compose_file is not None:
            yield self._compose_file
        else:
            compose_filename = f"docker-compose-{self._config.transport_mode}.yaml"
            with importlib.resources.path(__package__, compose_filename) as compose_file:
                yield compose_file

    def start(self) -> None:
        with self._get_compose_file() as compose_file:
            port_acp, port_ft = find_free_ports(2)

            if self._config.transport_mode == "mtls":
                assert self._config.certs_dir is not None
                self._transport_options = {
                    ServerKey.MAIN: MTLSOptions(
                        host="localhost",
                        port=port_acp,
                        certs_dir=pathlib.Path(self._config.certs_dir),
                        allow_remote_host=False,
                    ),
                    ServerKey.FILE_TRANSFER: MTLSOptions(
                        host="localhost",
                        port=port_ft,
                        certs_dir=pathlib.Path(self._config.certs_dir),
                        allow_remote_host=False,
                    ),
                }
            elif self._config.transport_mode == "insecure":
                self._transport_options = {
                    ServerKey.MAIN: InsecureOptions(
                        host="localhost",
                        port=port_acp,
                        allow_remote_host=False,
                    ),
                    ServerKey.FILE_TRANSFER: InsecureOptions(
                        host="localhost",
                        port=port_ft,
                        allow_remote_host=False,
                    ),
                }
            else:
                raise ValueError(
                    f"Unsupported transport mode '{self._config.transport_mode}'. "
                    "Only 'mtls' and 'insecure' are supported."
                )

            env = collections.ChainMap(
                {"PORT_ACP": str(port_acp), "PORT_FILETRANSFER": str(port_ft)}, self._env
            )

            # The compose_file may be temporary, in particular if the package is a zipfile.
            # To avoid it being deleted before docker compose has read it, we use the '--wait'
            # flag for 'docker compose'.
            cmd = self._compose_cmds + [
                "-f",
                str(compose_file.resolve()),
                "--project-name",
                self._compose_name,
                "up",
                "--detach",
            ]
            if self._compose_version >= parse_version("2.1.1"):
                # The '--wait' flag is only available from version >= 2.1.1 of docker compose:
                # https://github.com/docker/compose/commit/72e4519cbfb6cdfc600e6ebfa377ce4b8e162c78
                cmd.append("--wait")
            proc_res = subprocess.run(  # nosec B603: documented in 'security_considerations.rst'
                cmd, env=env, capture_output=True, text=True
            )
            if proc_res.returncode != 0:
                raise RuntimeError(f"Docker compose failed to start:\n{proc_res.stderr}")

    def stop(self, *, timeout: float | None = None) -> None:
        # The compose file needs to be passed for all commands with docker-compose 1.X.
        # With docker-compose 2.X, this no longer seems to be necessary.
        with self._get_compose_file() as compose_file:
            cmd = self._compose_cmds + [
                "-f",
                str(compose_file),
                "--project-name",
                self._compose_name,
                "down",
            ]
            if timeout is not None:
                # --timeout must be an integer, so we round up.
                cmd.extend(["--timeout", str(math.ceil(timeout))])
            if not self._keep_volume:
                cmd.append("--volumes")
            subprocess.check_call(  # nosec B603: documented in 'security_considerations.rst'
                cmd,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )

    def check(self, timeout: float | None = None) -> bool:
        for transport_options in self.transport_options.values():
            channel = transport_options.create_channel()
            if not check_grpc_health(channel=channel, timeout=timeout):
                return False
        return True

    @property
    def transport_options(self) -> dict[str, TransportOptionsType]:  # type: ignore[override]
        return self._transport_options
